{
 "cells": [
  {
   "cell_type": "markdown",
   "id": "26d2c6fc-6368-4335-9a2c-1faa75ff4ece",
   "metadata": {},
   "source": [
    "# Intro to classes\n",
    "\n",
    "Your first step into intermediate Python begins with classes. Classes are at the core of Python: everything is an object in Python, which means everything has a class. Even built-in objects written in C are still Python objects with classes."
   ]
  },
  {
   "cell_type": "markdown",
   "id": "474071b5-839b-4328-b7ce-3ddfeada6bf3",
   "metadata": {},
   "source": [
    "## What is an object?\n",
    "\n",
    "An object is simply a collection of data and functions that operate on that data.\n",
    "\n",
    "For example, let's say we wanted to represent our home directory as an object. It might look something like this:\n",
    "\n",
    "```\n",
    "home_directory:\n",
    "  string_location = \"/home/me\"\n",
    "  exists(self) -> bool\n",
    "```\n",
    "\n",
    "This object holds a single data \"member\" (`string_location`), and has a function, called a \"method\", to see if the directory exists.\n",
    "\n",
    "We could produce lots of these, each with different `string_location` values, and we could use them in our code to track directories and see if they exist. All of these objects are interchangeable, and all of them have identical functions - only the contents of the data are different. This suggests we could make a further improvement to the model. (Unless we were in JavaScript, by the way, where this really _is_ how objects were implemented!)"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "a99596a7-c949-4c3d-b1ac-27f16fb1abc9",
   "metadata": {},
   "source": [
    "## What is a class?\n",
    "\n",
    "Now we will make a \"template\" for creating new objects; called a class."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "23d3aaeb-5f9b-4ecc-8604-c4b846e17d97",
   "metadata": {},
   "outputs": [],
   "source": [
    "import os\n",
    "\n",
    "\n",
    "class Path:\n",
    "    def __init__(self, string_location):\n",
    "        self.string_location = string_location\n",
    "\n",
    "    def exists(self):\n",
    "        return os.path.exists(self.string_location)"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "a396914b-764c-48e1-bf8a-cf3cfe10cfeb",
   "metadata": {},
   "outputs": [],
   "source": [
    "usr_bin = Path(\"/usr/bin\")\n",
    "usr_bin.exists()"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "aa5e34a8-75e8-415d-bac0-e60318ef9eae",
   "metadata": {},
   "source": [
    "The `__init__` method is special to Python: if you \"call\" the class, Python will create a new instance of the class, then call it's `__init__` method, passing in the new, empty instance into \"self\". Inside this method, we add `string_location` to `self`."
   ]
  },
  {
   "cell_type": "markdown",
   "id": "3078983b-f817-4434-aae2-37d9cc7c2e18",
   "metadata": {},
   "source": [
    "Notice that Python automatically knows that calling a method on a class instance should pass the instance as the first arguments. We could have written this instead, which is identical:"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "054af9d8-6171-4b27-bd82-f1cb58f3eded",
   "metadata": {},
   "outputs": [],
   "source": [
    "Path.exists(usr_bin)"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "089888da-844f-4ce2-92bc-c1d4cf20c0da",
   "metadata": {},
   "source": [
    "But it's a lot more convenient and concise to call it on the instance itself."
   ]
  },
  {
   "cell_type": "markdown",
   "id": "501a35e5-f61f-4184-97c1-0e0a833d41f1",
   "metadata": {},
   "source": [
    "The thing called `usr_bin` only caries the data we've assigned to it:"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "a462be59-96bb-4484-8b6e-1a8a3f2af9eb",
   "metadata": {},
   "outputs": [],
   "source": [
    "usr_bin.__dict__"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "b73033ca-b610-43fc-9cca-038eb7736239",
   "metadata": {},
   "source": [
    "It remembers its class, though:"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "6f82c7cf-283e-40c6-94f2-7755a51c63c8",
   "metadata": {},
   "outputs": [],
   "source": [
    "usr_bin.__class__"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "a2025612-a9e6-4ff6-bbef-344ec50683cd",
   "metadata": {},
   "source": [
    "When you try to access attributes, it checks, the object first, then it tries the class:"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "3fe69d31-6bfd-4679-b6f5-f84aa31fa6ad",
   "metadata": {},
   "outputs": [],
   "source": [
    "usr_bin.__class__.__dict__"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "c3a5208c-13e5-4bf5-8805-ddd205ae9221",
   "metadata": {},
   "source": [
    "There's some autogenerated stuff in there, but you can see `exists` is there too!"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "e2003ba6-77bd-4ffb-ad52-256aae38dafe",
   "metadata": {},
   "source": [
    "### Advanced: subclassing"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "205e0bf4-ff50-4d87-90ec-fa476069b2cf",
   "metadata": {},
   "source": [
    "Why stop there? It's often useful to organize in further levels. This is accomplished by subclassing - a class can be \"based on\" another class - what this means is that the most specific class is checked first, but then it keeps going up the chain. This lookup mechanism is called the mro (method resolution order), and you can check it explicitly:"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "49fcd227-e0bf-4096-90c5-9ce62db1530d",
   "metadata": {},
   "outputs": [],
   "source": [
    "usr_bin.__class__.__mro__"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "39f751e9-2952-4f00-b9fe-c5ddca6e20f9",
   "metadata": {},
   "source": [
    "All classes are eventually subclasses of `object` - the last item in this list. That's where the default behaviors come from."
   ]
  },
  {
   "cell_type": "markdown",
   "id": "5880afb3-8550-4fe3-9304-95b02ef7571c",
   "metadata": {},
   "source": [
    "## Simpler: dataclasses\n",
    "\n",
    "If you come from a compiled language, the syntax for making a class might be unusual for you. You might be more used to simply listing the members and methods together, something like this:\n",
    "\n",
    "\n",
    "```python\n",
    "class BadPath:\n",
    "    string_location = ...\n",
    "\n",
    "    def exists(self):\n",
    "        return os.path.exists(self.string_location)\n",
    "```\n",
    "\n",
    "\n",
    "Question: Why is this wrong?\n",
    "\n",
    "Answer: This member variable is _on the class_. That means all `BadPath`'s would have the same `string_location`! We also don't know what to assign it too (it should be assigned when you make an instance for an instance variable).\n",
    "\n",
    "\n",
    "But... Wouldn't it be nice if we didn't have to be so repetitive? Well, we can have the best of both worlds:"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "ee0237cd-1cc8-4a94-8a03-3b98255384c9",
   "metadata": {},
   "outputs": [],
   "source": [
    "import dataclasses\n",
    "\n",
    "\n",
    "@dataclasses.dataclass\n",
    "class DataPath:\n",
    "    string_location: str\n",
    "\n",
    "    def exists(self):\n",
    "        return os.path.exists(self.string_location)"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "d5793cbf-9435-49b6-9986-bda69b6322ff",
   "metadata": {},
   "source": [
    "We just add a decorator (we'll cover these later, for now it's just a marker that processes this class into the correct output. In fact, that's kind of always what they are). We also add a type annotation since Python doesn't allow a variable declaration without at least a type annotation or a value. The type annotation just tells the reader (not Python) what type this expects. Python still doesn't care what you really do with this."
   ]
  },
  {
   "cell_type": "markdown",
   "id": "5c550062-e742-43bb-abe3-3aca08edc68f",
   "metadata": {},
   "source": [
    "Now, we get an `__init__` for free!"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "a044dd61-83db-4779-abd3-d97978dda232",
   "metadata": {},
   "outputs": [],
   "source": [
    "DataPath(\"/usr/local\")"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "a8c5a843-53af-4e7c-8833-d9b66e38fb95",
   "metadata": {},
   "source": [
    "In fact, we got a lot more for free. Notice now nicely it printed out? Compare that to our old class:"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "efa6f95a-8eec-448d-8970-f922acb03afe",
   "metadata": {},
   "outputs": [],
   "source": [
    "Path(\"/usr/local\")"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "f0532ebf-8f82-4134-b58d-a623d8258285",
   "metadata": {},
   "source": [
    "That's the default `object` repr, which just tells you the `__class__.__name__` and memory location (ugh), instead of something more helpful. We would have had to do a lot more work to make a nice class with the vanilla syntax!"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "eb78bbd1-3e04-4a39-9880-40b9285cd83d",
   "metadata": {},
   "source": [
    "There are a lot of useful options in dataclasses that can help you make useful classes; here are most of them:\n",
    "\n",
    "- `init`: Make a init function (default: True)\n",
    "- `repr`: Make a nice repr (default: True)\n",
    "- `eq`: Allow equality (default: True)\n",
    "- `order`: Allow comparisons (default: False)\n",
    "- `frozen`: Disallow mutation (default: False)\n",
    "- `slots`: Keep the class from accepting new members (Python 3.10+, default: False, slots classes have no `__dict__`)\n",
    "- `kw_only`: Do not allow pass-by position (Python 3.10, default False, frees up subclassing a lot)\n",
    "- `match_args`: Support Python 3.10 pattern matching via position (Python 3.10, default: True)\n",
    "\n",
    "You can also control each attribute (field in dataclass terms) with options, and you can specify `__post_init__`, which runs after the generated `__init__`.\n",
    "\n",
    "If you like dataclasses, feel free to check out `attrs`, which inspired dataclasses and is a little more powerful, `cattrs`, which handles conversions for both the stdlib dataclasses and attrs, as well as `pydantic`, which is an all-in-one solution for data conversion and validation too, but less flexible."
   ]
  },
  {
   "cell_type": "markdown",
   "id": "7464ff6b-e8f9-46cb-a587-41c8faddfac9",
   "metadata": {},
   "source": [
    "## Using classes"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "478f4796-44ea-4097-b37a-13e56c637568",
   "metadata": {},
   "source": [
    "Let's look at a built-in class, `int`:"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "abd7fc01-73de-4996-b2b2-a83e902cf170",
   "metadata": {},
   "outputs": [],
   "source": [
    "my_int = int(3)"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "26ee85ce-d455-4ce9-9c66-eef5815fd169",
   "metadata": {},
   "source": [
    "Since this is so common, there's a built in shortcut for this - we could have used `my_int = 3` directly - Python turns numbers into integers when it sees them. We can call methods, too:"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "bd9234d1-b2a9-4526-8207-8d15665f9831",
   "metadata": {},
   "outputs": [],
   "source": [
    "my_int.bit_length()"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "298108a5-965f-4943-baf8-42899e0dd328",
   "metadata": {},
   "source": [
    "It takes 2 bits to be able to represent this integer. Python uses many more than that, but this is useful information about integers.\n",
    "\n",
    "> Note: you cannot write `3.bit_length()`; due to the Python parser, this is invalid syntax due to Python thinking you've started writing a float. You can, however, do this with a float. `2.0.is_integer()` is valid, for example. As is `(2).bit_length()`."
   ]
  },
  {
   "cell_type": "markdown",
   "id": "e48fe023-c3a1-458a-8019-0688de5367fd",
   "metadata": {},
   "source": [
    "## Special methods\n",
    "\n",
    "We can't go very far without writing a special method - `__init__` we've already seen. Python has a lot of special methods that have double underscores before and after the name - called \"dunder methods\". These customize all sorts of things about the class. Let's try to make our \"plain\" class look more like our \"dataclass\":"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "c90dbacc-eb8f-4e8d-a605-3a8fefb8c09c",
   "metadata": {},
   "outputs": [],
   "source": [
    "class Path:\n",
    "    def __init__(self, string_location):\n",
    "        self.string_location = string_location\n",
    "\n",
    "    def exists(self):\n",
    "        return os.path.exists(self.string_location)\n",
    "\n",
    "    def __repr__(self):\n",
    "        return f\"{self.__class__.__name__}(string_location={self.string_location!r})\""
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "0e623a06-264d-4164-9302-21f7f18107f5",
   "metadata": {},
   "outputs": [],
   "source": [
    "Path(\"/usr/local\")"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "9b8ad967-a069-40d6-9f55-87e252d43711",
   "metadata": {},
   "source": [
    "This looks like our `DataPath` now! We've customized what the \"representation\" of the object of this class look like. We could also separately control what the string representation (`__str__`) look like, which will allow the printed form and the REPL form to look different, which is a really nice feature of Python missing from some other languages like Matlab. `repr` is usually programmer friendly, and `str` is usually user friendly."
   ]
  },
  {
   "cell_type": "markdown",
   "id": "38619c31-ee6b-464c-a492-7c4b7a24019e",
   "metadata": {},
   "source": [
    "You can also control what most operators do on the class, like comparison:"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "8a5b6874-7fb0-4417-98c4-2888bf0b6a1a",
   "metadata": {},
   "outputs": [],
   "source": [
    "DataPath(\"/usr/local\") == DataPath(\"/usr/local\")"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "ecaec26a-7ce1-493e-8de0-1989a7799798",
   "metadata": {},
   "outputs": [],
   "source": [
    "Path(\"/usr/local\") == Path(\"/usr/local\")"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "6c2f987e-cb9e-4338-aa59-c7a0b81df33c",
   "metadata": {},
   "source": [
    "Yeah, like `__repr__`, dataclasses generated a reasonable default `__eq__` method for us, while the vanilla class just falls back on `object`'s `__eq__`, which checks to see if the objects share the same memory (which these do not).\n",
    "\n",
    "We could add this manually. Let's use inheritance to add it, since we are ~~lazy~~ good programmers and don't like repeating ourselves:"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "851975ab-81a9-4ea8-a058-9c7ed12a1617",
   "metadata": {},
   "outputs": [],
   "source": [
    "class EqPath(Path):\n",
    "    def __eq__(self, other):\n",
    "        return self.string_location == other.string_location"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "e07e9c21-4614-4bc3-9a38-2f8f0cbf0dd0",
   "metadata": {},
   "outputs": [],
   "source": [
    "EqPath(\"/usr/local\") == EqPath(\"/usr/local\")"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "3c969b1c-3e35-49e4-8f73-e8dd4892b33d",
   "metadata": {},
   "source": [
    "What if we wanted to sort paths alphabetically? Neither of our class families support it out of the box:"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "b4579c60-bf67-439c-a379-c0e7192b77d3",
   "metadata": {},
   "outputs": [],
   "source": [
    "# sorted([EqPath(\"/loc/a\"), EqPath(\"/loc/b\")])\n",
    "# sorted([DataPath(\"/loc/a\"), DataPath(\"/loc/b\")])"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "a0079558-222b-46c5-893e-c73b3a1e7753",
   "metadata": {},
   "source": [
    "Adding this would require adding `__lt__`, or dataclasses to the rescue again with `order=True`):"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "2f9cb863-7db2-4a48-a3a4-ae97cdadc035",
   "metadata": {},
   "outputs": [],
   "source": [
    "import dataclasses\n",
    "\n",
    "\n",
    "@dataclasses.dataclass(order=True)\n",
    "class DataPath:\n",
    "    string_location: str\n",
    "\n",
    "    def exists(self):\n",
    "        return os.path.exists(self.string_location)"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "20cdba99-1779-4da7-9fb4-ad3f7e78b481",
   "metadata": {},
   "outputs": [],
   "source": [
    "print(*sorted([DataPath(\"/loc/b\"), DataPath(\"/loc/a\")]), sep=\"\\n\")"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "aaad6e69-bc9e-4966-a73e-32982eeba54a",
   "metadata": {},
   "source": [
    "How does this work? It treats these like a tuple when sorting: the first field is sorted first."
   ]
  }
 ],
 "metadata": {
  "kernelspec": {
   "display_name": "Python [conda env:level-up-your-python] *",
   "language": "python",
   "name": "conda-env-level-up-your-python-py"
  },
  "language_info": {
   "codemirror_mode": {
    "name": "ipython",
    "version": 3
   },
   "file_extension": ".py",
   "mimetype": "text/x-python",
   "name": "python",
   "nbconvert_exporter": "python",
   "pygments_lexer": "ipython3",
   "version": "3.10.5"
  }
 },
 "nbformat": 4,
 "nbformat_minor": 5
}
